# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/internal_spark_cluster_status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/internal_spark_cluster_status.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n5tecton_proto/data/internal_spark_cluster_status.proto\x12\x11tecton_proto.data\"\xc3\x01\n\x1aInternalSparkClusterStatus\x12I\n\x06status\x18\x01 \x01(\x0e\x32\x31.tecton_proto.data.InternalSparkClusterStatusEnumR\x06status\x12\x14\n\x05\x65rror\x18\x02 \x01(\x08R\x05\x65rror\x12#\n\rerror_message\x18\x03 \x01(\tR\x0c\x65rrorMessage\x12\x1f\n\x0b\x63luster_url\x18\x04 \x01(\tR\nclusterUrl*\xc5\x02\n\x1eInternalSparkClusterStatusEnum\x12)\n%INTERNAL_SPARK_CLUSTER_STATUS_UNKNOWN\x10\x00\x12,\n(INTERNAL_SPARK_CLUSTER_STATUS_NO_CLUSTER\x10\x01\x12\x32\n.INTERNAL_SPARK_CLUSTER_STATUS_CREATING_CLUSTER\x10\x02\x12>\n:INTERNAL_SPARK_CLUSTER_STATUS_WAITING_FOR_CLUSTER_TO_START\x10\x03\x12)\n%INTERNAL_SPARK_CLUSTER_STATUS_HEALTHY\x10\x04\x12+\n\'INTERNAL_SPARK_CLUSTER_STATUS_UNHEALTHY\x10\x05\x42\x13\n\x0f\x63om.tecton.dataP\x01'
)

_INTERNALSPARKCLUSTERSTATUSENUM = _descriptor.EnumDescriptor(
  name='InternalSparkClusterStatusEnum',
  full_name='tecton_proto.data.InternalSparkClusterStatusEnum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_SPARK_CLUSTER_STATUS_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_SPARK_CLUSTER_STATUS_NO_CLUSTER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_SPARK_CLUSTER_STATUS_CREATING_CLUSTER', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_SPARK_CLUSTER_STATUS_WAITING_FOR_CLUSTER_TO_START', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_SPARK_CLUSTER_STATUS_HEALTHY', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_SPARK_CLUSTER_STATUS_UNHEALTHY', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=275,
  serialized_end=600,
)
_sym_db.RegisterEnumDescriptor(_INTERNALSPARKCLUSTERSTATUSENUM)

InternalSparkClusterStatusEnum = enum_type_wrapper.EnumTypeWrapper(_INTERNALSPARKCLUSTERSTATUSENUM)
INTERNAL_SPARK_CLUSTER_STATUS_UNKNOWN = 0
INTERNAL_SPARK_CLUSTER_STATUS_NO_CLUSTER = 1
INTERNAL_SPARK_CLUSTER_STATUS_CREATING_CLUSTER = 2
INTERNAL_SPARK_CLUSTER_STATUS_WAITING_FOR_CLUSTER_TO_START = 3
INTERNAL_SPARK_CLUSTER_STATUS_HEALTHY = 4
INTERNAL_SPARK_CLUSTER_STATUS_UNHEALTHY = 5



_INTERNALSPARKCLUSTERSTATUS = _descriptor.Descriptor(
  name='InternalSparkClusterStatus',
  full_name='tecton_proto.data.InternalSparkClusterStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='tecton_proto.data.InternalSparkClusterStatus.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='status', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='tecton_proto.data.InternalSparkClusterStatus.error', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='error', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='tecton_proto.data.InternalSparkClusterStatus.error_message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errorMessage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cluster_url', full_name='tecton_proto.data.InternalSparkClusterStatus.cluster_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='clusterUrl', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=77,
  serialized_end=272,
)

_INTERNALSPARKCLUSTERSTATUS.fields_by_name['status'].enum_type = _INTERNALSPARKCLUSTERSTATUSENUM
DESCRIPTOR.message_types_by_name['InternalSparkClusterStatus'] = _INTERNALSPARKCLUSTERSTATUS
DESCRIPTOR.enum_types_by_name['InternalSparkClusterStatusEnum'] = _INTERNALSPARKCLUSTERSTATUSENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

InternalSparkClusterStatus = _reflection.GeneratedProtocolMessageType('InternalSparkClusterStatus', (_message.Message,), {
  'DESCRIPTOR' : _INTERNALSPARKCLUSTERSTATUS,
  '__module__' : 'tecton_proto.data.internal_spark_cluster_status_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.InternalSparkClusterStatus)
  })
_sym_db.RegisterMessage(InternalSparkClusterStatus)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



try:
    import grpc
except ImportError:
    raise ImportError("Please install grpcio")

from tecton._internals.metadata_service_impl.grpc import grpc_stub
from tecton_spark.logger import get_logger


logger = get_logger("grpc_over_grpc")


def channel():
    channel_options = [
        ("grpc.max_message_length", 64 * 1024 * 1024),
        ("grpc.max_receive_message_length", 64 * 1024 * 1024),
    ]
    return grpc.insecure_channel(grpc_stub._get_host_port(), options=channel_options)



import warnings
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import pendulum
from google.protobuf.duration_pb2 import Duration
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType
from pyspark.sql.types import TimestampType
from typeguard import typechecked

from tecton._internals import errors
from tecton._internals.feature_views.aggregations import construct_full_tafv_df
from tecton.declarative.base import BaseDataSource
from tecton.declarative.transformation import Transformation
from tecton.run_api_consts import AGGREGATION_LEVEL_DISABLED
from tecton.run_api_consts import AGGREGATION_LEVEL_FULL
from tecton.run_api_consts import AGGREGATION_LEVEL_PARTIAL
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME
from tecton.run_api_consts import SUPPORTED_AGGREGATION_LEVEL_VALUES
from tecton_proto.args.feature_view_pb2 import AggregationMode as AggregationModeProto
from tecton_proto.args.feature_view_pb2 import FeatureAggregation as FeatureAggregationProto
from tecton_proto.args.feature_view_pb2 import FeatureViewArgs
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_proto.common.aggregation_function_pb2 import AggregationFunction
from tecton_proto.data.feature_store_pb2 import FeatureStoreFormatVersion
from tecton_proto.data.feature_view_pb2 import AggregateFeature
from tecton_proto.data.feature_view_pb2 import TrailingTimeWindowAggregation
from tecton_proto.data.new_transformation_pb2 import NewTransformation as TransformationProto
from tecton_proto.data.virtual_data_source_pb2 import VirtualDataSource as VirtualDataSourceProto
from tecton_spark import time_utils
from tecton_spark.id_helper import IdHelper
from tecton_spark.partial_aggregations import construct_partial_time_aggregation_df
from tecton_spark.partial_aggregations import rename_partial_aggregate_columns
from tecton_spark.pipeline_helper import get_all_input_keys
from tecton_spark.pipeline_helper import pipeline_to_dataframe


@typechecked
def declarative_run(
    args: FeatureViewArgs,
    spark: SparkSession,
    start_time: Optional[datetime],
    end_time: Optional[datetime],
    aggregation_level: Optional[str],
    mock_sources: Dict[str, DataFrame],
) -> DataFrame:
    """
    Internal implementation of run() for materialized (i.e. not on-demand) feature view declarative objects.
    """
    assert args.HasField(
        "materialized_feature_view_args"
    ), "declarative_run can only be used with materialized feature views"
    if aggregation_level and not args.materialized_feature_view_args.aggregations:
        raise errors.FV_UNSUPPORTED_ARG("aggregation_level")

    # Set default aggregation_level.
    if not aggregation_level:
        aggregation_level = (
            AGGREGATION_LEVEL_FULL if args.materialized_feature_view_args.aggregations else AGGREGATION_LEVEL_DISABLED
        )

    if aggregation_level not in SUPPORTED_AGGREGATION_LEVEL_VALUES:
        raise errors.FV_INVALID_ARG_VALUE(
            "aggregation_level", str(aggregation_level), str(SUPPORTED_AGGREGATION_LEVEL_VALUES)
        )

    _validate_batch_mock_sources_keys(args, mock_sources)

    batch_schedule = _get_batch_schedule(args)
    start_time, end_time = _resolve_times(batch_schedule, start_time, end_time)
    _warn_incorrect_time_range_size(args, start_time, end_time, aggregation_level)

    df = pipeline_to_dataframe(
        spark,
        pipeline=args.pipeline,
        consume_streaming_data_sources=False,
        data_sources=_get_data_sources(args),
        transformations=get_transformations(args),
        feature_time_limits=pendulum.period(start_time, end_time),
        schedule_interval=batch_schedule,
        mock_inputs=mock_sources,
    )
    timestamp_field = _get_timestamp_field(args, df.schema)
    df = df.filter((df[timestamp_field] >= start_time) & (df[timestamp_field] < end_time))
    if aggregation_level == AGGREGATION_LEVEL_DISABLED:
        return df

    trailing_time_window_aggregation = _construct_trailing_time_window_aggregation(args, df.schema)
    join_keys = [join_key for entity in args.entities for join_key in entity.join_keys]

    if aggregation_level == AGGREGATION_LEVEL_PARTIAL:
        # Perform partial rollup with human readable output format.
        df = construct_partial_time_aggregation_df(
            df=df,
            join_keys=join_keys,
            time_aggregation=trailing_time_window_aggregation,
            version=FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TTL_FIELD,
            window_start_column_name=DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME,
            window_end_column_name=DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME,
            convert_to_epoch=False,
            aggregation_anchor_time=end_time,
        )
        aggregation_interval_string = _duration_to_human_readable_string(
            args.materialized_feature_view_args.aggregation_interval
        )
        return rename_partial_aggregate_columns(
            df=df,
            slide_interval_string=aggregation_interval_string,
            trailing_time_window_aggregation=trailing_time_window_aggregation,
        )

    # Perform partial rollup for each aggregate tile.
    df = construct_partial_time_aggregation_df(
        df=df,
        join_keys=join_keys,
        time_aggregation=trailing_time_window_aggregation,
        version=FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TTL_FIELD,
        aggregation_anchor_time=end_time,
    )

    # Perform final rollup from aggregate tiles up to each result window.
    return construct_full_tafv_df(
        spark=spark,
        time_aggregation=trailing_time_window_aggregation,
        join_keys=join_keys,
        feature_store_format_version=FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TTL_FIELD,
        tile_interval=time_utils.proto_to_duration(args.materialized_feature_view_args.aggregation_interval),
        all_partial_aggregations_df=df,
        use_materialized_data=False,
    )


@typechecked
def _warn_incorrect_time_range_size(
    args: FeatureViewArgs, start_time: datetime, end_time: datetime, aggregation_level: Optional[str]
):
    time_range = end_time - start_time
    batch_schedule = _get_batch_schedule(args)
    if args.materialized_feature_view_args.aggregations:
        aggregation_interval = args.materialized_feature_view_args.aggregation_interval.ToTimedelta()
        if aggregation_level == AGGREGATION_LEVEL_FULL:
            max_aggregation = max(
                (agg.time_window.ToTimedelta() for agg in args.materialized_feature_view_args.aggregations)
            )
            if time_range < max_aggregation:
                # Uses warnings.warn() instead of tecton_spark.logger.warning() because it works better with pytest.
                warnings.warn(
                    f"Run time range ({start_time}, {end_time}) is smaller than the maximum aggregation size: {max_aggregation}. This may lead to incorrect aggregate feature values."
                )
            if time_range.total_seconds() % aggregation_interval.total_seconds() != 0:
                warnings.warn(
                    f"Run time range ({start_time}, {end_time}) is not a multiple of the aggregation_interval: {aggregation_interval}. This may lead to incorrect aggregate feature values, since Tecton pre-aggregates data in smaller time windows based on the aggregation_interval size."
                )
        elif aggregation_level == AGGREGATION_LEVEL_PARTIAL:
            if time_range.total_seconds() % aggregation_interval.total_seconds() != 0:
                warnings.warn(
                    f"Run time range ({start_time}, {end_time}) is not a multiple of the aggregation_interval: {aggregation_interval}. This may lead to incorrect aggregate feature values, since Tecton pre-aggregates data in smaller time windows based on the aggregation_interval size."
                )
    elif args.materialized_feature_view_args.incremental_backfills and time_range != batch_schedule:
        warnings.warn(
            f"Run time range ({start_time}, {end_time}) is not equivalent to the batch_schedule: {batch_schedule}. This may lead to incorrect feature values since feature views with incremental_backfills typically implicitly rely on the materialization range being equivalent to the batch_schedule."
        )


@typechecked
def _get_data_sources(args: FeatureViewArgs) -> List[VirtualDataSourceProto]:
    """
    Returns a list of all of the data source data protos that this feature view depends on.

    Note: Normally data protos are computed in metadata server but not in this case.
    """
    from tecton._internals.fco import _ALL_FCOS

    ds_ids = _get_ids_from_pipeline(args.pipeline.root, "data_source_node")

    data_source_data_protos = []
    for ds_id in ds_ids:
        ds = _ALL_FCOS[ds_id]
        data_source_data_protos.append(_data_proto_from_data_source(ds))

    return data_source_data_protos


@typechecked
def _data_proto_from_data_source(ds: BaseDataSource) -> VirtualDataSourceProto:
    data_proto = VirtualDataSourceProto()
    data_proto.virtual_data_source_id.CopyFrom(ds._id)
    data_proto.fco_metadata.name = ds._args.info.name
    data_proto.batch_data_source.timestamp_column_properties.column_name = ds.timestamp_field
    return data_proto


@typechecked
def get_transformations(args: FeatureViewArgs) -> List[TransformationProto]:
    """
    Returns a list of all of the transformation data protos that this feature view depends on.

    Note: Normally data protos are computed in metadata server but not in this case.
    """
    from tecton._internals.fco import _ALL_FCOS

    transformation_ids = _get_ids_from_pipeline(args.pipeline.root, "transformation_node")

    transformation_data_protos = []
    for transformation_id in transformation_ids:
        transformation = _ALL_FCOS[transformation_id]
        transformation_data_protos.append(_data_proto_from_transformation(transformation))

    return transformation_data_protos


@typechecked
def _data_proto_from_transformation(transformation: Transformation) -> TransformationProto:
    data_proto = TransformationProto()
    data_proto.transformation_id.CopyFrom(transformation._id)
    data_proto.fco_metadata.name = transformation._args.info.name
    data_proto.transformation_mode = transformation._args.transformation_mode
    data_proto.user_function.CopyFrom(transformation._args.user_function)
    return data_proto


@typechecked
def _get_ids_from_pipeline(node: PipelineNode, desired_node_type: str) -> Set[str]:
    """
    Returns the FCO ids of all ancestor pipeline nodes of the desired node type.
    """
    assert desired_node_type in ("data_source_node", "transformation_node"), "Unsupported desired_node_type"
    ids = set()
    if node.HasField("data_source_node") and desired_node_type == "data_source_node":
        ids.add(IdHelper.to_string(node.data_source_node.virtual_data_source_id))
    elif node.HasField("transformation_node"):
        if desired_node_type == "transformation_node":
            ids.add(IdHelper.to_string(node.transformation_node.transformation_id))
        for child in node.transformation_node.inputs:
            ids.update(_get_ids_from_pipeline(child.node, desired_node_type))
    return ids


@typechecked
def _resolve_times(
    batch_schedule: timedelta,
    start_time: Optional[datetime],
    end_time: Optional[datetime],
) -> Tuple[datetime, datetime]:
    assert start_time or end_time, "At least one of start_time or end_time should be set."
    # If start_time or end_time time is missing, default to one batch_schedule.
    if start_time is None:
        start_time = end_time - batch_schedule
    elif end_time is None:
        end_time = start_time + batch_schedule

    if end_time <= start_time:
        raise errors.END_TIME_BEFORE_START_TIME(start_time, end_time)

    return start_time, end_time


@typechecked
def _construct_trailing_time_window_aggregation(
    args: FeatureViewArgs, df_schema: StructType
) -> TrailingTimeWindowAggregation:
    aggregation = TrailingTimeWindowAggregation()
    aggregation.time_key = _get_timestamp_field(args, df_schema)
    aggregation.is_continuous = (
        args.materialized_feature_view_args.aggregation_mode == AggregationModeProto.AGGREGATION_MODE_CONTINUOUS
    )
    aggregation.aggregation_slide_period.CopyFrom(args.materialized_feature_view_args.aggregation_interval)

    for feature_aggregation in args.materialized_feature_view_args.aggregations:
        aggregation.features.extend(
            _create_aggregate_features(feature_aggregation, args.materialized_feature_view_args.aggregation_interval)
        )
    return aggregation


@typechecked
def _create_aggregate_features(
    feature_aggregation: FeatureAggregationProto, aggregation_interval_seconds: Duration
) -> List[AggregateFeature]:
    """Build a list of AggregateFeature from the input FeatureAggregationProto."""
    aggregation_features = []
    feature_function = AggregationFunction.Value(f"AGGREGATION_FUNCTION_{feature_aggregation.function.upper()}")
    feature = AggregateFeature()
    feature.input_feature_name = feature_aggregation.column
    feature.function = feature_function
    feature.window.CopyFrom(feature_aggregation.time_window)

    if feature.function == AggregationFunction.AGGREGATION_FUNCTION_LASTN:
        feature.function_params.last_n.n = feature_aggregation.function_params["n"].int64_value
        function_name = feature_aggregation.function.lower() + str(feature.function_params.last_n.n)
    else:
        function_name = feature_aggregation.function.lower()
    feature.output_feature_name = feature_aggregation.name or _construct_output_feature_name(
        column=feature_aggregation.column,
        function=function_name,
        window=feature_aggregation.time_window,
        aggregation_interval=aggregation_interval_seconds,
    )
    aggregation_features.append(feature)
    return aggregation_features


@typechecked
def _construct_output_feature_name(column: str, function: str, window: Duration, aggregation_interval: Duration):
    window_name = _duration_to_human_readable_string(window)
    aggregation_interval_name = _duration_to_human_readable_string(aggregation_interval)
    return f"{column}_{function}_{window_name}_{aggregation_interval_name}".replace(" ", "")


@typechecked
def _get_batch_schedule(args: FeatureViewArgs) -> timedelta:
    assert args.HasField("materialized_feature_view_args")
    batch_schedule = args.materialized_feature_view_args.batch_schedule.ToTimedelta()
    if batch_schedule > timedelta(0):
        return batch_schedule

    aggregation_interval = args.materialized_feature_view_args.aggregation_interval.ToTimedelta()
    assert aggregation_interval > timedelta(0), "batch_schedule or aggregation_interval must be > 0"
    return aggregation_interval


@typechecked
def _get_timestamp_field(args: FeatureViewArgs, df_schema: StructType) -> str:
    if args.materialized_feature_view_args.timestamp_field:
        return args.materialized_feature_view_args.timestamp_field

    timestamp_fields = [field for field in df_schema if field.dataType == TimestampType()]

    assert (
        len(timestamp_fields) == 1
    ), "To use run(), timestamp_field must be set or the feature view output should contain exactly one TimestampType column."
    return timestamp_fields[0].name


@typechecked
def _validate_batch_mock_sources_keys(args: FeatureViewArgs, mock_sources: Dict[str, DataFrame]):
    expected_input_names = get_all_input_keys(args.pipeline.root)
    mock_sources_keys = set(mock_sources.keys())
    if not mock_sources_keys.issubset(expected_input_names):
        raise errors.FV_INVALID_MOCK_SOURCES(mock_sources_keys, expected_input_names)


@typechecked
def _duration_to_human_readable_string(duration: Duration) -> str:
    """
    Replicate the backend logic that converts durations to feature strings, e.g. `column_sum_1h30m_30m`.

    Backend logic: https://github.com/tecton-ai/tecton/blob/e3c04e28066c21fa4ed9b459ece458c084766b6e/java/com/tecton/common/utils/TimeUtils.kt#L80
    """
    duration = duration.ToTimedelta()

    # Timedelta's internal representation is a reduced days, seconds, and microseconds.
    days = duration.days
    hours = int(duration.seconds / 3600)
    minutes = int((duration.seconds % 3600) / 60)
    seconds = duration.seconds % 60

    string = ""
    if days:
        string += f"{days}d"
    if hours:
        string += f"{hours}h"
    if minutes:
        string += f"{minutes}m"
    if seconds:
        string += f"{seconds}s"
    if not string:
        string = "0s"  # Continuous

    return string



from typing import Optional


# NOTE: When adding a new format, change data_source_helper.py:_partition_value_for_time if needed
class DatetimePartitionColumn:
    """
    Helper class to tell Tecton how underlying flat files are date/time partitioned for Hive/Glue data sources. This can translate into a significant performance increase.

    You will generally include an object of this class in the `datetime_partition_columns` option in a `HiveConfig` object.
    """

    # pass zero_padded=True if the partition column is a string that is zero-padded
    def __init__(self, column_name, datepart: str, zero_padded: bool = False, format_string: Optional[str] = None):
        """
        Instantiates a new DatetimePartitionColumn configuration object.


        :param column_name: The name of the column in the Glue/Hive schema that corresponds to the underlying date/time partition folder. Note that if you do not explicitly specify a name in your partition folders, Glue will name the column of the form ``partition_0``.
        :param datepart: The part of the date that this column specifies. Can be one of "year", "month", "day", "hour", or the full "date". If used with ``format_string``, this should be the size of partition being represented, e.g. ``datepart="month"`` for ``format_string="%Y-%m"``.
        :param zero_padded: Whether the ``datepart`` has a leading zero if less than two digits. This must be set to True if ``datepart="date"``. (Should not be set if ``format_string`` is set.)
        :param format_string:
            A ``datetime.strftime`` format string override for "non-default" partition columns formats. E.g. ``"%Y%m%d"`` for ``datepart="date"`` instead of the Tecton default ``"%Y-%m-%d"``, or ``"%Y-%m"`` for ``datepart="month"`` instead of the Tecton default ``"%m"``.

            IMPORTANT: This format string must convert python datetimes (via ``datetime.strftime(format)``) to strings that are sortable in time order. For example, ``"%m-%Y"`` would be an invalid format string because ``"09-2019" > "05-2020"``.

            See https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes for format codes.

        Example definitions:

            Assume you have an S3 bucket with parquet files stored in the following structure: ``s3://mybucket/2022/05/04/<multiple parquet files>`` , where ``2022`` is the year, ``05`` is the month, and ``04`` is the day of the month. In this scenario, you could use the following definition:

            .. code-block:: python

                datetime_partition_columns = [
                    DatetimePartitionColumn(column_name="partition_0", datepart="year", zero_padded=True),
                    DatetimePartitionColumn(column_name="partition_1", datepart="month", zero_padded=True),
                    DatetimePartitionColumn(column_name="partition_2", datepart="day", zero_padded=True),
                ]

            Example using the ``format_string`` parameter. Assume your data is partitioned by ``"YYYY-MM"``, e.g. ``s3://mybucket/2022-05/<multiple parquet files>``. Tecton's default month format is ``"%m"``, which would fail to format datetime strings that are comparable to your table's partition column, so the definition needs to specify an override.

            .. code-block:: python

                datetime_partition_columns = [
                    DatetimePartitionColumn(column_name="partition_1", datepart="month", format_string="%Y-%m"),
                ]

        :return: DatetimePartitionColumn instantiation
        """
        self.column_name = column_name
        self.datepart = datepart
        self.zero_padded = zero_padded
        self.format_string = format_string



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/state_update.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.args import fco_args_pb2 as tecton__proto_dot_args_dot_fco__args__pb2
from tecton_proto.args import repo_metadata_pb2 as tecton__proto_dot_args_dot_repo__metadata__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/state_update.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n$tecton_proto/data/state_update.proto\x12\x11tecton_proto.data\x1a\x1ctecton_proto/common/id.proto\x1a tecton_proto/args/fco_args.proto\x1a%tecton_proto/args/repo_metadata.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a$tecton_proto/args/diff_options.proto\"\xd0\x01\n\x0f\x46\x63oPropertyDiff\x12#\n\rproperty_name\x18\x01 \x01(\tR\x0cpropertyName\x12!\n\x0cval_existing\x18\x02 \x01(\tR\x0bvalExisting\x12!\n\x0cval_declared\x18\x03 \x01(\tR\x0bvalDeclared\x12R\n\x0erendering_type\x18\x04 \x01(\x0e\x32+.tecton_proto.args.FcoPropertyRenderingTypeR\rrenderingType\"\xe6\x01\n\x1c\x42\x61tchMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12\x46\n\x11schedule_interval\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x10scheduleInterval\x12W\n\x0b\x64\x65stination\x18\x03 \x01(\x0e\x32\x35.tecton_proto.data.MaterializationTaskDiffDestinationR\x0b\x64\x65stination\"F\n\x1dStreamMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\"\xd7\x02\n\x1f\x42\x61\x63kfillMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12H\n\x12\x66\x65\x61ture_start_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12$\n\x0enumber_of_jobs\x18\x04 \x01(\x05R\x0cnumberOfJobs\x12W\n\x0b\x64\x65stination\x18\x05 \x01(\x0e\x32\x35.tecton_proto.data.MaterializationTaskDiffDestinationR\x0b\x64\x65stination\"\xae\x02\n\x13MaterializationInfo\x12\x62\n\x13\x62\x61\x63kfill_task_diffs\x18\x01 \x03(\x0b\x32\x32.tecton_proto.data.BackfillMaterializationTaskDiffR\x11\x62\x61\x63kfillTaskDiffs\x12W\n\x0f\x62\x61tch_task_diff\x18\x02 \x01(\x0b\x32/.tecton_proto.data.BatchMaterializationTaskDiffR\rbatchTaskDiff\x12Z\n\x10stream_task_diff\x18\x03 \x01(\x0b\x32\x30.tecton_proto.data.StreamMaterializationTaskDiffR\x0estreamTaskDiff\"\xde\x02\n\x07\x46\x63oDiff\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x04type\x12?\n\rexisting_args\x18\x07 \x01(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x0c\x65xistingArgs\x12?\n\rdeclared_args\x18\x08 \x01(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x0c\x64\x65\x63laredArgs\x12\x36\n\x04\x64iff\x18\x04 \x03(\x0b\x32\".tecton_proto.data.FcoPropertyDiffR\x04\x64iff\x12Y\n\x14materialization_info\x18\x0b \x01(\x0b\x32&.tecton_proto.data.MaterializationInfoR\x13materializationInfoJ\x04\x08\t\x10\n\"C\n\x0b\x46\x63oFieldRef\x12.\n\x06\x66\x63o_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x05\x66\x63oIdJ\x04\x08\x02\x10\x03\"h\n\x11ValidationMessage\x12\x18\n\x07message\x18\x01 \x01(\tR\x07message\x12\x39\n\x08\x66\x63o_refs\x18\x02 \x03(\x0b\x32\x1e.tecton_proto.data.FcoFieldRefR\x07\x66\x63oRefs\"\x92\x01\n\x10ValidationResult\x12<\n\x06\x65rrors\x18\x01 \x03(\x0b\x32$.tecton_proto.data.ValidationMessageR\x06\x65rrors\x12@\n\x08warnings\x18\x02 \x03(\x0b\x32$.tecton_proto.data.ValidationMessageR\x08warnings\"\xe3\x02\n\x12StateUpdateRequest\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12\x35\n\x08\x66\x63o_args\x18\x02 \x03(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x07\x66\x63oArgs\x12R\n\x10repo_source_info\x18\x03 \x01(\x0b\x32(.tecton_proto.args.FeatureRepoSourceInfoR\x0erepoSourceInfo\x12-\n\x12suppress_recreates\x18\n \x01(\x08R\x11suppressRecreates\x12\x1f\n\x0bupgrade_all\x18\x04 \x01(\x08R\nupgradeAll\x12!\n\x0crequested_by\x18\x05 \x01(\tR\x0brequestedBy\x12\x1f\n\x0bsdk_version\x18\t \x01(\tR\nsdkVersionJ\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\t\"\xa7\x01\n\x10StateUpdateEntry\x12\x1b\n\tcommit_id\x18\x01 \x01(\tR\x08\x63ommitId\x12\x1d\n\napplied_by\x18\x02 \x01(\tR\tappliedBy\x12\x39\n\napplied_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tappliedAt\x12\x1c\n\tworkspace\x18\x04 \x01(\tR\tworkspace\"\xb8\x02\n\x16StateUpdatePlanSummary\x12L\n\ndiff_items\x18\x01 \x03(\x0b\x32-.tecton_proto.data.StateUpdatePlanSummaryDiffR\tdiffItems\x12\x1d\n\napplied_by\x18\x02 \x01(\tR\tappliedBy\x12\x1d\n\ncreated_by\x18\x03 \x01(\tR\tcreatedBy\x12\x39\n\napplied_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tappliedAt\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x1c\n\tworkspace\x18\x06 \x01(\tR\tworkspace\"\xbc\x02\n\x1aStateUpdatePlanSummaryDiff\x12\x19\n\x08\x66\x63o_type\x18\x01 \x01(\tR\x07\x66\x63oType\x12\x38\n\x04type\x18\x02 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x04type\x12\x38\n\x05\x64iffs\x18\x03 \x03(\x0b\x32\".tecton_proto.data.FcoPropertyDiffR\x05\x64iffs\x12\x12\n\x04name\x18\x04 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x05 \x01(\tR\x0b\x64\x65scription\x12Y\n\x14materialization_info\x18\x06 \x01(\x0b\x32&.tecton_proto.data.MaterializationInfoR\x13materializationInfo*n\n\x11\x46\x63oTransitionType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x43REATE\x10\x01\x12\n\n\x06\x44\x45LETE\x10\x02\x12\x0b\n\x07UPGRADE\x10\x05\x12\n\n\x06UPDATE\x10\x03\x12\x0c\n\x08RECREATE\x10\x04\x12\r\n\tUNCHANGED\x10\x06*\xfe\x01\n\"MaterializationTaskDiffDestination\x12\x35\n1MATERIALIZATION_TASK_DIFF_DESTINATION_UNSPECIFIED\x10\x00\x12\x31\n-MATERIALIZATION_TASK_DIFF_DESTINATION_OFFLINE\x10\x01\x12\x30\n,MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE\x10\x02\x12<\n8MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE_AND_OFFLINE\x10\x03\x42\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_fco__args__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_repo__metadata__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,tecton__proto_dot_args_dot_diff__options__pb2.DESCRIPTOR,])

_FCOTRANSITIONTYPE = _descriptor.EnumDescriptor(
  name='FcoTransitionType',
  full_name='tecton_proto.data.FcoTransitionType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DELETE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPGRADE', index=3, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPDATE', index=4, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RECREATE', index=5, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNCHANGED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=3271,
  serialized_end=3381,
)
_sym_db.RegisterEnumDescriptor(_FCOTRANSITIONTYPE)

FcoTransitionType = enum_type_wrapper.EnumTypeWrapper(_FCOTRANSITIONTYPE)
_MATERIALIZATIONTASKDIFFDESTINATION = _descriptor.EnumDescriptor(
  name='MaterializationTaskDiffDestination',
  full_name='tecton_proto.data.MaterializationTaskDiffDestination',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_TASK_DIFF_DESTINATION_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_TASK_DIFF_DESTINATION_OFFLINE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE_AND_OFFLINE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=3384,
  serialized_end=3638,
)
_sym_db.RegisterEnumDescriptor(_MATERIALIZATIONTASKDIFFDESTINATION)

MaterializationTaskDiffDestination = enum_type_wrapper.EnumTypeWrapper(_MATERIALIZATIONTASKDIFFDESTINATION)
UNKNOWN = 0
CREATE = 1
DELETE = 2
UPGRADE = 5
UPDATE = 3
RECREATE = 4
UNCHANGED = 6
MATERIALIZATION_TASK_DIFF_DESTINATION_UNSPECIFIED = 0
MATERIALIZATION_TASK_DIFF_DESTINATION_OFFLINE = 1
MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE = 2
MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE_AND_OFFLINE = 3



_FCOPROPERTYDIFF = _descriptor.Descriptor(
  name='FcoPropertyDiff',
  full_name='tecton_proto.data.FcoPropertyDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='property_name', full_name='tecton_proto.data.FcoPropertyDiff.property_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='propertyName', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='val_existing', full_name='tecton_proto.data.FcoPropertyDiff.val_existing', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='valExisting', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='val_declared', full_name='tecton_proto.data.FcoPropertyDiff.val_declared', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='valDeclared', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rendering_type', full_name='tecton_proto.data.FcoPropertyDiff.rendering_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='renderingType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=266,
  serialized_end=474,
)


_BATCHMATERIALIZATIONTASKDIFF = _descriptor.Descriptor(
  name='BatchMaterializationTaskDiff',
  full_name='tecton_proto.data.BatchMaterializationTaskDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_string', full_name='tecton_proto.data.BatchMaterializationTaskDiff.display_string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='displayString', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schedule_interval', full_name='tecton_proto.data.BatchMaterializationTaskDiff.schedule_interval', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='scheduleInterval', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination', full_name='tecton_proto.data.BatchMaterializationTaskDiff.destination', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='destination', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=477,
  serialized_end=707,
)


_STREAMMATERIALIZATIONTASKDIFF = _descriptor.Descriptor(
  name='StreamMaterializationTaskDiff',
  full_name='tecton_proto.data.StreamMaterializationTaskDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_string', full_name='tecton_proto.data.StreamMaterializationTaskDiff.display_string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='displayString', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=709,
  serialized_end=779,
)


_BACKFILLMATERIALIZATIONTASKDIFF = _descriptor.Descriptor(
  name='BackfillMaterializationTaskDiff',
  full_name='tecton_proto.data.BackfillMaterializationTaskDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_string', full_name='tecton_proto.data.BackfillMaterializationTaskDiff.display_string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='displayString', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_start_time', full_name='tecton_proto.data.BackfillMaterializationTaskDiff.feature_start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureStartTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_end_time', full_name='tecton_proto.data.BackfillMaterializationTaskDiff.feature_end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureEndTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='number_of_jobs', full_name='tecton_proto.data.BackfillMaterializationTaskDiff.number_of_jobs', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='numberOfJobs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination', full_name='tecton_proto.data.BackfillMaterializationTaskDiff.destination', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='destination', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=782,
  serialized_end=1125,
)


_MATERIALIZATIONINFO = _descriptor.Descriptor(
  name='MaterializationInfo',
  full_name='tecton_proto.data.MaterializationInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='backfill_task_diffs', full_name='tecton_proto.data.MaterializationInfo.backfill_task_diffs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='backfillTaskDiffs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_task_diff', full_name='tecton_proto.data.MaterializationInfo.batch_task_diff', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='batchTaskDiff', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stream_task_diff', full_name='tecton_proto.data.MaterializationInfo.stream_task_diff', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='streamTaskDiff', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1128,
  serialized_end=1430,
)


_FCODIFF = _descriptor.Descriptor(
  name='FcoDiff',
  full_name='tecton_proto.data.FcoDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='tecton_proto.data.FcoDiff.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='type', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='existing_args', full_name='tecton_proto.data.FcoDiff.existing_args', index=1,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='existingArgs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='declared_args', full_name='tecton_proto.data.FcoDiff.declared_args', index=2,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='declaredArgs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='diff', full_name='tecton_proto.data.FcoDiff.diff', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='diff', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_info', full_name='tecton_proto.data.FcoDiff.materialization_info', index=4,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationInfo', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1433,
  serialized_end=1783,
)


_FCOFIELDREF = _descriptor.Descriptor(
  name='FcoFieldRef',
  full_name='tecton_proto.data.FcoFieldRef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fco_id', full_name='tecton_proto.data.FcoFieldRef.fco_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcoId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1785,
  serialized_end=1852,
)


_VALIDATIONMESSAGE = _descriptor.Descriptor(
  name='ValidationMessage',
  full_name='tecton_proto.data.ValidationMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='tecton_proto.data.ValidationMessage.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='message', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fco_refs', full_name='tecton_proto.data.ValidationMessage.fco_refs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcoRefs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1854,
  serialized_end=1958,
)


_VALIDATIONRESULT = _descriptor.Descriptor(
  name='ValidationResult',
  full_name='tecton_proto.data.ValidationResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='errors', full_name='tecton_proto.data.ValidationResult.errors', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='errors', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warnings', full_name='tecton_proto.data.ValidationResult.warnings', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='warnings', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1961,
  serialized_end=2107,
)


_STATEUPDATEREQUEST = _descriptor.Descriptor(
  name='StateUpdateRequest',
  full_name='tecton_proto.data.StateUpdateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='workspace', full_name='tecton_proto.data.StateUpdateRequest.workspace', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workspace', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fco_args', full_name='tecton_proto.data.StateUpdateRequest.fco_args', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcoArgs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repo_source_info', full_name='tecton_proto.data.StateUpdateRequest.repo_source_info', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='repoSourceInfo', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='suppress_recreates', full_name='tecton_proto.data.StateUpdateRequest.suppress_recreates', index=3,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='suppressRecreates', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='upgrade_all', full_name='tecton_proto.data.StateUpdateRequest.upgrade_all', index=4,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='upgradeAll', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='requested_by', full_name='tecton_proto.data.StateUpdateRequest.requested_by', index=5,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='requestedBy', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sdk_version', full_name='tecton_proto.data.StateUpdateRequest.sdk_version', index=6,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sdkVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2110,
  serialized_end=2465,
)


_STATEUPDATEENTRY = _descriptor.Descriptor(
  name='StateUpdateEntry',
  full_name='tecton_proto.data.StateUpdateEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit_id', full_name='tecton_proto.data.StateUpdateEntry.commit_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='commitId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='applied_by', full_name='tecton_proto.data.StateUpdateEntry.applied_by', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='appliedBy', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='applied_at', full_name='tecton_proto.data.StateUpdateEntry.applied_at', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='appliedAt', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='workspace', full_name='tecton_proto.data.StateUpdateEntry.workspace', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workspace', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2468,
  serialized_end=2635,
)


_STATEUPDATEPLANSUMMARY = _descriptor.Descriptor(
  name='StateUpdatePlanSummary',
  full_name='tecton_proto.data.StateUpdatePlanSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='diff_items', full_name='tecton_proto.data.StateUpdatePlanSummary.diff_items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='diffItems', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='applied_by', full_name='tecton_proto.data.StateUpdatePlanSummary.applied_by', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='appliedBy', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by', full_name='tecton_proto.data.StateUpdatePlanSummary.created_by', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='createdBy', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='applied_at', full_name='tecton_proto.data.StateUpdatePlanSummary.applied_at', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='appliedAt', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_at', full_name='tecton_proto.data.StateUpdatePlanSummary.created_at', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='createdAt', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='workspace', full_name='tecton_proto.data.StateUpdatePlanSummary.workspace', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workspace', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2638,
  serialized_end=2950,
)


_STATEUPDATEPLANSUMMARYDIFF = _descriptor.Descriptor(
  name='StateUpdatePlanSummaryDiff',
  full_name='tecton_proto.data.StateUpdatePlanSummaryDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fco_type', full_name='tecton_proto.data.StateUpdatePlanSummaryDiff.fco_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcoType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='tecton_proto.data.StateUpdatePlanSummaryDiff.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='type', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='diffs', full_name='tecton_proto.data.StateUpdatePlanSummaryDiff.diffs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='diffs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='tecton_proto.data.StateUpdatePlanSummaryDiff.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='tecton_proto.data.StateUpdatePlanSummaryDiff.description', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='description', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='materialization_info', full_name='tecton_proto.data.StateUpdatePlanSummaryDiff.materialization_info', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='materializationInfo', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2953,
  serialized_end=3269,
)

_FCOPROPERTYDIFF.fields_by_name['rendering_type'].enum_type = tecton__proto_dot_args_dot_diff__options__pb2._FCOPROPERTYRENDERINGTYPE
_BATCHMATERIALIZATIONTASKDIFF.fields_by_name['schedule_interval'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_BATCHMATERIALIZATIONTASKDIFF.fields_by_name['destination'].enum_type = _MATERIALIZATIONTASKDIFFDESTINATION
_BACKFILLMATERIALIZATIONTASKDIFF.fields_by_name['feature_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BACKFILLMATERIALIZATIONTASKDIFF.fields_by_name['feature_end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BACKFILLMATERIALIZATIONTASKDIFF.fields_by_name['destination'].enum_type = _MATERIALIZATIONTASKDIFFDESTINATION
_MATERIALIZATIONINFO.fields_by_name['backfill_task_diffs'].message_type = _BACKFILLMATERIALIZATIONTASKDIFF
_MATERIALIZATIONINFO.fields_by_name['batch_task_diff'].message_type = _BATCHMATERIALIZATIONTASKDIFF
_MATERIALIZATIONINFO.fields_by_name['stream_task_diff'].message_type = _STREAMMATERIALIZATIONTASKDIFF
_FCODIFF.fields_by_name['type'].enum_type = _FCOTRANSITIONTYPE
_FCODIFF.fields_by_name['existing_args'].message_type = tecton__proto_dot_args_dot_fco__args__pb2._FCOARGS
_FCODIFF.fields_by_name['declared_args'].message_type = tecton__proto_dot_args_dot_fco__args__pb2._FCOARGS
_FCODIFF.fields_by_name['diff'].message_type = _FCOPROPERTYDIFF
_FCODIFF.fields_by_name['materialization_info'].message_type = _MATERIALIZATIONINFO
_FCOFIELDREF.fields_by_name['fco_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_VALIDATIONMESSAGE.fields_by_name['fco_refs'].message_type = _FCOFIELDREF
_VALIDATIONRESULT.fields_by_name['errors'].message_type = _VALIDATIONMESSAGE
_VALIDATIONRESULT.fields_by_name['warnings'].message_type = _VALIDATIONMESSAGE
_STATEUPDATEREQUEST.fields_by_name['fco_args'].message_type = tecton__proto_dot_args_dot_fco__args__pb2._FCOARGS
_STATEUPDATEREQUEST.fields_by_name['repo_source_info'].message_type = tecton__proto_dot_args_dot_repo__metadata__pb2._FEATUREREPOSOURCEINFO
_STATEUPDATEENTRY.fields_by_name['applied_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_STATEUPDATEPLANSUMMARY.fields_by_name['diff_items'].message_type = _STATEUPDATEPLANSUMMARYDIFF
_STATEUPDATEPLANSUMMARY.fields_by_name['applied_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_STATEUPDATEPLANSUMMARY.fields_by_name['created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_STATEUPDATEPLANSUMMARYDIFF.fields_by_name['type'].enum_type = _FCOTRANSITIONTYPE
_STATEUPDATEPLANSUMMARYDIFF.fields_by_name['diffs'].message_type = _FCOPROPERTYDIFF
_STATEUPDATEPLANSUMMARYDIFF.fields_by_name['materialization_info'].message_type = _MATERIALIZATIONINFO
DESCRIPTOR.message_types_by_name['FcoPropertyDiff'] = _FCOPROPERTYDIFF
DESCRIPTOR.message_types_by_name['BatchMaterializationTaskDiff'] = _BATCHMATERIALIZATIONTASKDIFF
DESCRIPTOR.message_types_by_name['StreamMaterializationTaskDiff'] = _STREAMMATERIALIZATIONTASKDIFF
DESCRIPTOR.message_types_by_name['BackfillMaterializationTaskDiff'] = _BACKFILLMATERIALIZATIONTASKDIFF
DESCRIPTOR.message_types_by_name['MaterializationInfo'] = _MATERIALIZATIONINFO
DESCRIPTOR.message_types_by_name['FcoDiff'] = _FCODIFF
DESCRIPTOR.message_types_by_name['FcoFieldRef'] = _FCOFIELDREF
DESCRIPTOR.message_types_by_name['ValidationMessage'] = _VALIDATIONMESSAGE
DESCRIPTOR.message_types_by_name['ValidationResult'] = _VALIDATIONRESULT
DESCRIPTOR.message_types_by_name['StateUpdateRequest'] = _STATEUPDATEREQUEST
DESCRIPTOR.message_types_by_name['StateUpdateEntry'] = _STATEUPDATEENTRY
DESCRIPTOR.message_types_by_name['StateUpdatePlanSummary'] = _STATEUPDATEPLANSUMMARY
DESCRIPTOR.message_types_by_name['StateUpdatePlanSummaryDiff'] = _STATEUPDATEPLANSUMMARYDIFF
DESCRIPTOR.enum_types_by_name['FcoTransitionType'] = _FCOTRANSITIONTYPE
DESCRIPTOR.enum_types_by_name['MaterializationTaskDiffDestination'] = _MATERIALIZATIONTASKDIFFDESTINATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FcoPropertyDiff = _reflection.GeneratedProtocolMessageType('FcoPropertyDiff', (_message.Message,), {
  'DESCRIPTOR' : _FCOPROPERTYDIFF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FcoPropertyDiff)
  })
_sym_db.RegisterMessage(FcoPropertyDiff)

BatchMaterializationTaskDiff = _reflection.GeneratedProtocolMessageType('BatchMaterializationTaskDiff', (_message.Message,), {
  'DESCRIPTOR' : _BATCHMATERIALIZATIONTASKDIFF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.BatchMaterializationTaskDiff)
  })
_sym_db.RegisterMessage(BatchMaterializationTaskDiff)

StreamMaterializationTaskDiff = _reflection.GeneratedProtocolMessageType('StreamMaterializationTaskDiff', (_message.Message,), {
  'DESCRIPTOR' : _STREAMMATERIALIZATIONTASKDIFF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.StreamMaterializationTaskDiff)
  })
_sym_db.RegisterMessage(StreamMaterializationTaskDiff)

BackfillMaterializationTaskDiff = _reflection.GeneratedProtocolMessageType('BackfillMaterializationTaskDiff', (_message.Message,), {
  'DESCRIPTOR' : _BACKFILLMATERIALIZATIONTASKDIFF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.BackfillMaterializationTaskDiff)
  })
_sym_db.RegisterMessage(BackfillMaterializationTaskDiff)

MaterializationInfo = _reflection.GeneratedProtocolMessageType('MaterializationInfo', (_message.Message,), {
  'DESCRIPTOR' : _MATERIALIZATIONINFO,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.MaterializationInfo)
  })
_sym_db.RegisterMessage(MaterializationInfo)

FcoDiff = _reflection.GeneratedProtocolMessageType('FcoDiff', (_message.Message,), {
  'DESCRIPTOR' : _FCODIFF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FcoDiff)
  })
_sym_db.RegisterMessage(FcoDiff)

FcoFieldRef = _reflection.GeneratedProtocolMessageType('FcoFieldRef', (_message.Message,), {
  'DESCRIPTOR' : _FCOFIELDREF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.FcoFieldRef)
  })
_sym_db.RegisterMessage(FcoFieldRef)

ValidationMessage = _reflection.GeneratedProtocolMessageType('ValidationMessage', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATIONMESSAGE,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.ValidationMessage)
  })
_sym_db.RegisterMessage(ValidationMessage)

ValidationResult = _reflection.GeneratedProtocolMessageType('ValidationResult', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATIONRESULT,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.ValidationResult)
  })
_sym_db.RegisterMessage(ValidationResult)

StateUpdateRequest = _reflection.GeneratedProtocolMessageType('StateUpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _STATEUPDATEREQUEST,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.StateUpdateRequest)
  })
_sym_db.RegisterMessage(StateUpdateRequest)

StateUpdateEntry = _reflection.GeneratedProtocolMessageType('StateUpdateEntry', (_message.Message,), {
  'DESCRIPTOR' : _STATEUPDATEENTRY,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.StateUpdateEntry)
  })
_sym_db.RegisterMessage(StateUpdateEntry)

StateUpdatePlanSummary = _reflection.GeneratedProtocolMessageType('StateUpdatePlanSummary', (_message.Message,), {
  'DESCRIPTOR' : _STATEUPDATEPLANSUMMARY,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.StateUpdatePlanSummary)
  })
_sym_db.RegisterMessage(StateUpdatePlanSummary)

StateUpdatePlanSummaryDiff = _reflection.GeneratedProtocolMessageType('StateUpdatePlanSummaryDiff', (_message.Message,), {
  'DESCRIPTOR' : _STATEUPDATEPLANSUMMARYDIFF,
  '__module__' : 'tecton_proto.data.state_update_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.StateUpdatePlanSummaryDiff)
  })
_sym_db.RegisterMessage(StateUpdatePlanSummaryDiff)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from typing import List
from typing import Optional
from typing import Union

import attr
import pendulum
import pytimeparse
from google.protobuf.duration_pb2 import Duration

from tecton._internals import errors
from tecton.aggregation_functions import AggregationFunction
from tecton_proto.args import feature_view_pb2


BACKFILL_CONFIG_MODE_PROTO_PREFIX = "BACKFILL_CONFIG_MODE_"
BACKFILL_CONFIG_MODE_MULTIPLE = "multiple_batch_schedule_intervals_per_job"
BACKFILL_CONFIG_MODE_SINGLE = "single_batch_schedule_interval_per_job"


@attr.s(auto_attribs=True)
class FeatureAggregation(object):
    """
    This class describes a single aggregation that is applied in a batch or stream window aggregate feature view.

    :param column: Column name of the feature we are aggregating.
    :type column: str
    :param function: One of the built-in aggregation functions.
    :type function: Union[str, AggregationFunction]
    :param time_windows: Duration to aggregate over in pytimeparse_ format. Examples: ``"30days"``, ``["8hours", "30days", "365days"]``.
    :type time_windows: Union[str, List[str]]

    `function` can be one of predefined numeric aggregation functions, namely ``"count"``, ``"sum"``, ``"mean"``, ``"min"``, ``"max"``. For
    these numeric aggregations, you can pass the name of it as a string. Nulls are handled like Spark SQL `Function(column)`, e.g. SUM/MEAN/MIN/MAX of all nulls is null and COUNT of all nulls is 0.

    In addition to numeric aggregations, :class:`FeatureAggregation` supports "last-n" aggregations that
    will compute the last N distinct values for the column by timestamp. Right now only string column types are supported as inputs
    to this aggregation, i.e., the resulting feature value will be a list of strings. Nulls are not included in the aggregated list.

    You can use it via the ``last_distinct()`` helper function like this:

    .. code-block:: python

        from tecton.aggregation_functions import last_distinct
        my_fv = BatchWindowAggregateFeatureView(
        ...
        aggregations=[FeatureAggregation(
            column='my_column',
            function=last_distinct(15),
            time_windows=['7days'])],
        ...
        )

    .. _pytimeparse: https://pypi.org/project/pytimeparse/
    """

    column: str
    """Column name of the feature we are aggregating."""
    function: Union[str, AggregationFunction]
    """One of the built-in aggregation functions (`'count'`, `'sum'`, `'mean'`, `'min'`, `'max'`)."""
    time_windows: Union[str, List[str]]
    """
       Examples: `"30days"`, `["8hours", "30days", "365days"]`.

       .. _pytimeparse: https://pypi.org/project/pytimeparse/
       """

    def _to_proto(self):
        proto = feature_view_pb2.FeatureAggregation()
        proto.column = self.column

        if isinstance(self.function, str):
            proto.function = self.function
        elif isinstance(self.function, AggregationFunction):
            proto.function = self.function.name
            for k, v in self.function.params.items():
                assert isinstance(v, int)
                proto.function_params[k].CopyFrom(feature_view_pb2.ParamValue(int64_value=v))
        else:
            raise TypeError(f"Invalid function type: {type(self.function)}")

        windows = self.time_windows if isinstance(self.time_windows, list) else [self.time_windows]
        for w in windows:
            duration = Duration()
            duration.FromTimedelta(pendulum.duration(seconds=pytimeparse.parse(w)))
            proto.time_windows.append(duration)
            proto.time_window_strs.append(w)
        return proto


@attr.s(auto_attribs=True)
class BackfillConfig:
    """Configuration used to specify backfill options.

    This class configures the backfill behavior of a Batch Feature View. Requires
    materialization to be enabled.

    :param mode: Determines whether Tecton batches backfill jobs:
        [``single_batch_schedule_interval_per_job``, ``multiple_batch_schedule_intervals_per_job``]
    """

    mode: str

    def _to_proto(self) -> feature_view_pb2.BackfillConfig:
        proto = feature_view_pb2.BackfillConfig()
        try:
            proto.mode = feature_view_pb2.BackfillConfigMode.Value(
                BACKFILL_CONFIG_MODE_PROTO_PREFIX + self.mode.upper()
            )
        except ValueError:
            # TODO: add BACKFILL_CONFIG_MODE_SINGLE when supported
            raise errors.InvalidBackfillConfigMode(self.mode, [BACKFILL_CONFIG_MODE_MULTIPLE])
        return proto


@attr.s(auto_attribs=True)
class MonitoringConfig:
    """Configuration used to specify monitoring options.

    This class describes the FeatureView materialization freshness and alerting configurations. Requires
    materialization to be enabled. Freshness monitoring requires online materialization to be enabled.
    See `Monitoring Materialization`_ for more details.

    :param monitor_freshness: Defines the enabled/disabled state of monitoring when feature data is materialized to the online feature store.
    :type monitor_freshness: bool
    :param expected_feature_freshness: Threshold used to determine if recently materialized feature data is stale.
        Data is stale if ``now - anchor_time(most_recent_feature_value) > expected_feature_freshness``.
        Value must be at least 2 times the feature tile length.
        If not specified, a value determined by the Tecton backend is used
    :type expected_feature_freshness: str, optional
    :param alert_email: Email that alerts for this FeatureView will be sent to.
    :type alert_email: str, optional

    An example declaration of a MonitorConfig

        .. code-block:: python

            from tecton import batch_feature_view, Input, MonitoringConfig
            # For all named arguments to the batch feature view, see docs for details and types.
            @batch_feature_view(
                sources=[FilteredSource(credit_scores_batch)],
                # Can be an argument instance to a batch feature view decorator
                monitoring = MonitoringConfig(
                    monitor_freshness=True,
                    expected_feature_freshness="1w",
                    alert_email="brian@tecton.ai"
                ),
                # Other named arguments
                ...
            )

            # Your batch feature view function
            def credit_batch_feature_view(credit_scores):
              ...

    .. _Monitoring Materialization: https://docs.tecton.ai/v2/overviews/monitoring_materialization.html
    """

    monitor_freshness: bool
    expected_feature_freshness: Optional[str] = None
    alert_email: Optional[str] = None

    def _to_proto(self) -> feature_view_pb2.MonitoringConfig:
        proto = feature_view_pb2.MonitoringConfig()

        if self.expected_feature_freshness:
            proto.expected_feature_freshness.FromTimedelta(
                pendulum.duration(seconds=pytimeparse.parse(self.expected_feature_freshness))
            )

        proto.alert_email = self.alert_email or ""
        proto.monitor_freshness = self.monitor_freshness
        return proto



from typing import Dict

from tecton._internals.fco import Fco
from tecton_proto.args.feature_view_pb2 import FeatureViewArgs
from tecton_proto.args.repo_metadata_pb2 import SourceInfo
from tecton_proto.common.id_pb2 import Id


class FeatureDefinition(Fco):
    """
    Represents the base class for Declarative FeatureViews and FeatureTables
    """

    _args: FeatureViewArgs
    _source_info: SourceInfo

    @property
    def _id(self) -> Id:
        return self._args.feature_view_id

    @property
    def name(self) -> str:
        """
        Name of this Tecton Object.
        """
        return self._args.info.name

    def __getitem__(self, features):
        from tecton.feature_services.feature_service_args import FeaturesConfig

        return FeaturesConfig(feature_view=self, namespace=self.name, features=features)

    def with_name(self, namespace: str):
        """
        Used to rename a Feature View used in a Feature Service.

        .. code-block:: python

            from tecton import FeatureService

            # The feature view in this feature service will be named "new_named_feature_view" in training data dataframe
            # columns and other metadata.
            feature_service = FeatureService(
                name="feature_service",
                features=[
                    my_feature_view.with_name("new_named_feature_view")
                ],
            )

            # Here is a more sophisticated example. The join keys for this feature service will be "transaction_id",
            # "sender_id", and "recipient_id" and will contain three feature views named "transaction_features",
            # "sender_features", and "recipient_features".
            transaction_fraud_service = FeatureService(
                name="transaction_fraud_service",
                features=[
                    # Select a subset of features from a feature view.
                    transaction_features[["amount"]],

                    # Rename a feature view and/or rebind its join keys. In this example, we want user features for both the
                    # transaction sender and recipient, so include the feature view twice and bind it to two different feature
                    # service join keys.
                    user_features.with_name("sender_features").with_join_key_map({"user_id" : "sender_id"}),
                    user_features.with_name("recipient_features").with_join_key_map({"user_id" : "recipient_id"}),
                ],
            )
        """
        from tecton.feature_services.feature_service_args import FeaturesConfig

        return FeaturesConfig(feature_view=self).with_name(namespace)

    def with_join_key_map(self, join_key_map: Dict[str, str]):
        """
        Used to rebind join keys for a Feature View used in a Feature Service. The keys in `join_key_map` should be the feature view join keys, and the values should be the feature service overrides.

        .. code-block:: python

            from tecton import FeatureService

            # The join key for this feature service will be "feature_service_user_id".
            feature_service = FeatureService(
                name="feature_service",
                features=[
                    my_feature_view.with_join_key_map({"user_id" : "feature_service_user_id"}),
                ],
            )

            # Here is a more sophisticated example. The join keys for this feature service will be "transaction_id",
            # "sender_id", and "recipient_id" and will contain three feature views named "transaction_features",
            # "sender_features", and "recipient_features".
            transaction_fraud_service = FeatureService(
                name="transaction_fraud_service",
                features=[
                    # Select a subset of features from a feature view.
                    transaction_features[["amount"]],

                    # Rename a feature view and/or rebind its join keys. In this example, we want user features for both the
                    # transaction sender and recipient, so include the feature view twice and bind it to two different feature
                    # service join keys.
                    user_features.with_name("sender_features").with_join_key_map({"user_id" : "sender_id"}),
                    user_features.with_name("recipient_features").with_join_key_map({"user_id" : "recipient_id"}),
                ],
            )
        """
        from tecton.feature_services.feature_service_args import FeaturesConfig

        return FeaturesConfig(feature_view=self).with_join_key_map(join_key_map)



from typing import Dict
from typing import List
from typing import Optional

from tecton._internals.feature_definition import FeatureDefinition


class FeaturesConfig(object):
    """
    Configuration used to specify a list of features.

    By default, you can add all of the features in a FeatureView/FeatureTable to a FeatureService by passing
    the FeatureView/FeatureTable into the ``features`` parameter of a FeatureService.
    However, if you want to specify a subset, you can use this class.

    You can use the double-bracket notation ``my_feature_view[[<features>]]``
    as a short-hand for generating a FeaturesConfig from a FeatureView. This is the preferred way to select a subset of
    of the features contained in a FeatureView. As an example:

    .. highlight:: python
    .. code-block:: python

       from tecton import FeatureService
       from feature_repo.features import my_feature_view_1, my_feature_view_2

       my_feature_service = FeatureService(
           name='my_feature_service',
           features=[
               # Add all features from my_feature_view_1 to this FeatureService
               my_feature_view_1,
               # Add a single feature from my_feature_view_2, 'my_feature'
               my_feature_view_2[['my_feature']]
           ]
       )

    :param namespace: A namespace used to prefix the features joined from this FeatureView.
        By default, namespace is set to the FeatureView name.
    :param features: The subset of features to select from the FeatureView.
    :param override_join_keys: (advanced) map of spine join key to feature view join key overrides.
    :param feature_view: The Feature View.
    """

    def __init__(
        self,
        *,
        feature_view: FeatureDefinition,
        namespace: str = None,
        features: Optional[List[str]] = None,
        override_join_keys: Optional[Dict[str, str]] = None,
    ):
        self._fv = feature_view
        assert self._fv is not None, "The `feature_view` field must be set."
        self.namespace = namespace or self._fv.name
        self.features = features
        self.override_join_keys = override_join_keys
        self.id = self._fv._id

    def with_name(self, namespace: str) -> "FeaturesConfig":
        self.namespace = namespace
        return self

    def with_join_key_map(self, join_key_map: Dict[str, str]) -> "FeaturesConfig":
        # As of FWV5, we map from spine key to feature view key. For backwards compatibility with FWV3,
        # self.override_join_keys is a map from feature view key to spine key. So we need to flip the input dict.
        self.override_join_keys = {v: k for k, v in join_key_map.items()}
        return self



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/framework_version.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/common/framework_version.proto',
  package='tecton_proto.common',
  syntax='proto2',
  serialized_options=b'\n\021com.tecton.commonP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n+tecton_proto/common/framework_version.proto\x12\x13tecton_proto.common*M\n\x10\x46rameworkVersion\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04\x46WV3\x10\x03\x12\x08\n\x04\x46WV4\x10\x04\x12\x08\n\x04\x46WV5\x10\x05\"\x04\x08\x01\x10\x01\"\x04\x08\x02\x10\x02\x42\x15\n\x11\x63om.tecton.commonP\x01'
)

_FRAMEWORKVERSION = _descriptor.EnumDescriptor(
  name='FrameworkVersion',
  full_name='tecton_proto.common.FrameworkVersion',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FWV3', index=1, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FWV4', index=2, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FWV5', index=3, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=68,
  serialized_end=145,
)
_sym_db.RegisterEnumDescriptor(_FRAMEWORKVERSION)

FrameworkVersion = enum_type_wrapper.EnumTypeWrapper(_FRAMEWORKVERSION)
UNSPECIFIED = 0
FWV3 = 3
FWV4 = 4
FWV5 = 5


DESCRIPTOR.enum_types_by_name['FrameworkVersion'] = _FRAMEWORKVERSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from abc import ABC
from abc import abstractmethod
from datetime import timedelta
from typing import List

from tecton._internals.fco import Fco
from tecton_proto.args import feature_view_pb2
from tecton_proto.args import virtual_data_source_pb2
from tecton_proto.args.repo_metadata_pb2 import SourceInfo
from tecton_proto.args.virtual_data_source_pb2 import VirtualDataSourceArgs
from tecton_proto.common.id_pb2 import Id


class BaseBatchDSConfig(ABC):
    @abstractmethod
    def _merge_batch_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        pass


class BaseStreamDSConfig(ABC):
    @abstractmethod
    def _merge_stream_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        pass


class BaseBatchConfig(ABC):
    @property
    @abstractmethod
    def data_delay(self) -> timedelta:
        pass

    @abstractmethod
    def _merge_batch_args(self, data_source_args: virtual_data_source_pb2.VirtualDataSourceArgs):
        pass


class BaseDataSource(Fco):
    _args: VirtualDataSourceArgs
    _source_info: SourceInfo

    @property
    def _id(self) -> Id:
        return self._args.virtual_data_source_id

    @property
    def name(self) -> str:
        """
        The name of this DataSource.
        """
        return self._args.info.name


class RequestSourceBase(ABC):
    @property
    @abstractmethod
    def schema(self):
        pass


class BaseEntity(Fco):
    @property
    @abstractmethod
    def name(self) -> str:
        """
        Name of the entity.
        """
        pass

    @property
    @abstractmethod
    def join_keys(self) -> List[str]:
        """
        Join keys of the entity.
        """
        pass


class FWV5BaseDataSource(BaseDataSource):
    @property
    def timestamp_field(self) -> str:
        """
        The name of the timestamp column or key of this DataSource.
        """
        if self._args.HasField("hive_ds_config"):
            return self._args.hive_ds_config.common_args.timestamp_field
        if self._args.HasField("redshift_ds_config"):
            return self._args.redshift_ds_config.common_args.timestamp_field
        if self._args.HasField("snowflake_ds_config"):
            return self._args.snowflake_ds_config.common_args.timestamp_field
        if self._args.HasField("file_ds_config"):
            return self._args.file_ds_config.common_args.timestamp_field
        else:
            raise Exception(f"Unknown Data Source Type: {self.name}")


class OutputStream(ABC):
    @abstractmethod
    def _to_proto() -> feature_view_pb2.OutputStream:
        pass



from typing import List
from typing import Union

from pyspark.sql.types import StructType

from tecton.declarative.base import RequestSourceBase
from tecton.types import Field
from tecton_spark.id_helper import IdHelper


class RequestDataSource(RequestSourceBase):
    """
    Declare a ``RequestDataSource``, for using request-time data in an ``OnDemandFeatureView``.
    """

    def __init__(
        self,
        request_schema: Union[StructType, List[Field]],
    ):
        """
        Creates a new RequestDataSource

        :param request_schema: PySpark schema for the RequestDataSource inputs.

        Example of a RequestDataSource declaration:

        .. code-block:: python

            from tecton import RequestDataSource
            from pyspark.sql.types import DoubleType, StructType, StructField

            request_schema = StructType([StructField('amount', DoubleType())])
            transaction_request = RequestDataSource(request_schema=request_schema)
        """
        self.request_schema = request_schema
        self.id = IdHelper.from_string(IdHelper.generate_string_id())

    @property
    def schema(self):
        return self.request_schema



from tecton.compat.data_sources.data_source import BatchDataSource
from tecton.compat.data_sources.data_source import StreamDataSource
from tecton.compat.data_sources.file_data_source import FileDSConfig
from tecton.compat.data_sources.hive_data_source import HiveDSConfig
from tecton.compat.data_sources.kafka_data_source import KafkaDSConfig
from tecton.compat.data_sources.kinesis_data_source import KinesisDSConfig
from tecton.compat.data_sources.redshift_data_source import RedshiftDSConfig
from tecton.compat.data_sources.request_data_source import RequestDataSource
from tecton.compat.data_sources.snowflake_data_source import SnowflakeDSConfig
from tecton.compat.entities.entity import Entity
from tecton.compat.entities.entity import OverriddenEntity
from tecton.compat.feature_configs import BackfillConfig
from tecton.compat.feature_configs import FeatureAggregation
from tecton.compat.feature_configs import MonitoringConfig
from tecton.compat.feature_service import FeatureService
from tecton.compat.feature_table import FeatureTable
from tecton.compat.feature_views.feature_view import batch_feature_view
from tecton.compat.feature_views.feature_view import batch_window_aggregate_feature_view
from tecton.compat.feature_views.feature_view import on_demand_feature_view
from tecton.compat.feature_views.feature_view import stream_feature_view
from tecton.compat.feature_views.feature_view import stream_window_aggregate_feature_view
from tecton.compat.inputs import Input
from tecton.compat.legacy_tecton_objects import DataSourceConfig
from tecton.compat.legacy_tecton_objects import MaterializationConfig
from tecton.compat.legacy_tecton_objects import NewDatabricksClusterConfig
from tecton.compat.legacy_tecton_objects import NewEMRClusterConfig
from tecton.compat.legacy_tecton_objects import online_transformation
from tecton.compat.legacy_tecton_objects import OnlineFeaturePackage
from tecton.compat.legacy_tecton_objects import PushFeaturePackage
from tecton.compat.legacy_tecton_objects import pyspark_transformation
from tecton.compat.legacy_tecton_objects import sql_transformation
from tecton.compat.legacy_tecton_objects import TemporalAggregateFeaturePackage
from tecton.compat.legacy_tecton_objects import TemporalFeaturePackage
from tecton.compat.legacy_tecton_objects import VirtualDataSource
from tecton.compat.transformation import tecton_sliding_window
from tecton.compat.transformation import transformation
from tecton.declarative.datetime_partition_column import DatetimePartitionColumn
from tecton.declarative.transformation import const
from tecton.declarative.transformation import PANDAS_MODE
from tecton.declarative.transformation import PYSPARK_MODE
from tecton.declarative.transformation import SPARK_SQL_MODE
from tecton.feature_services.feature_service_args import FeaturesConfig
from tecton.feature_services.logging_config import LoggingConfig
from tecton.features_common.feature_configs import DatabricksClusterConfig
from tecton.features_common.feature_configs import DeltaConfig
from tecton.features_common.feature_configs import DynamoConfig
from tecton.features_common.feature_configs import EMRClusterConfig
from tecton.features_common.feature_configs import ExistingClusterConfig
from tecton.features_common.feature_configs import ParquetConfig
from tecton.features_common.feature_configs import RedisConfig
from tecton_spark.function_serialization import inlined
from tecton_spark.materialization_context import materialization_context
from tecton_spark.time_utils import WINDOW_UNBOUNDED_PRECEDING

__all__ = [
    "Entity",
    "OverriddenEntity",
    "BatchDataSource",
    "StreamDataSource",
    "FileDSConfig",
    "HiveDSConfig",
    "KafkaDSConfig",
    "KinesisDSConfig",
    "RedshiftDSConfig",
    "SnowflakeDSConfig",
    "FeatureAggregation",
    "FeatureTable",
    "Input",
    "batch_feature_view",
    "batch_window_aggregate_feature_view",
    "stream_feature_view",
    "stream_window_aggregate_feature_view",
    "on_demand_feature_view",
    "BackfillConfig",
    "inlined",
    "sql_transformation",
    "pyspark_transformation",
    "online_transformation",
    "DataSourceConfig",
    "VirtualDataSource",
    "TemporalFeaturePackage",
    "TemporalAggregateFeaturePackage",
    "OnlineFeaturePackage",
    "PushFeaturePackage",
    "MaterializationConfig",
    "NewEMRClusterConfig",
    "NewDatabricksClusterConfig",
    "RequestDataSource",
    "tecton_sliding_window",
    "transformation",
    "WINDOW_UNBOUNDED_PRECEDING",
    "FeatureService",
    "FeaturesConfig",
    "MonitoringConfig",
    "LoggingConfig",
    "DatetimePartitionColumn",
    "const",
    "ExistingClusterConfig",
    "DeltaConfig",
    "ParquetConfig",
    "DatabricksClusterConfig",
    "EMRClusterConfig",
    "RedisConfig",
    "DynamoConfig",
    "materialization_context",
    "SPARK_SQL_MODE",
    "PYSPARK_MODE",
    "PANDAS_MODE",
]



import base64
from collections import defaultdict
from typing import Any
from typing import Tuple

import requests
from google.protobuf.empty_pb2 import Empty

from tecton._internals.metadata_service_impl import base_stub
from tecton._internals.metadata_service_impl import error_lib
from tecton._internals.metadata_service_impl import request_lib
from tecton._internals.metadata_service_impl.response import MDSResponse
from tecton_proto.metadataservice import metadata_service_pb2
from tecton_spark.logger import get_logger

logger = get_logger("metadata_service")

requests_session = requests.Session()


class PureHTTPStub(base_stub.BaseStub):
    def __init__(self):
        """
        PureHTTPStub mimics the interface exposed by a gRPC client for MetadataService.
        Callers can invoke a method like stub.GetFeatureView and it's transformed for an HTTP
        request using _InternalHTTPStub in the __getattr__ method below.
        """
        self.stub_obj = _InternalHTTPStub()

    def __getattr__(self, method_name):
        """
        Transforms methods called directly on PureHTTPStub to _InternalHTTPStub requests.
        E.g. PureHTTPStub::SomeMethod(request) is transformed to _InternalHTTPStub::execute('SomeMethod', request).

        An AttributeError is raised if the method invoked does not match a MetadataService RPC method.
        """
        methods = [m.name for m in metadata_service_pb2.DESCRIPTOR.services_by_name["MetadataService"].methods]
        if method_name not in methods:
            raise AttributeError("Nonexistent MetadataService method: " + method_name)

        def method(request, timeout_sec: float = 300.0):
            return self.stub_obj.execute(method_name, request, timeout_sec)

        return method

    def close(self):
        pass


class _InternalHTTPStub:
    def __init__(self):
        pass

    def execute(self, method: str, request, timeout_sec: float):
        """
        :param method: gRPC method name.
        :param request: Request proto.

        :return: Response proto.
        """
        json_request = {}
        json_request["method"] = f"/tecton_proto.metadataservice.MetadataService/{method}"
        json_request["metadata"] = request_lib.request_headers()
        request_serializer, response_deserializer = self._get_serializers(method)

        json_request["request"] = base64.encodebytes(request_serializer(request)).decode("utf-8")

        request_url = request_lib.request_url()
        response = requests_session.post(request_url, json=json_request, timeout=timeout_sec)
        response.raise_for_status()
        body = response.json()

        code = body["status"]["code"]
        if code != error_lib.gRPCStatus.OK.value:
            details = body["status"]["detail"]
            error_lib.raise_for_grpc_status(code, details, request_url)

        response_bytes = base64.decodebytes(body["response"].encode("utf-8"))
        response_proto = response_deserializer(response_bytes)
        metadata = body.get("metadata", "")
        return MDSResponse(response_proto, defaultdict(str, metadata))

    def _get_serializers(self, method_name) -> Tuple[Any, Any]:
        """
        :param method_name: gRPC method name, e.g. GetFeatureView.

        :return: A tuple of request serializer and request deserializer objects.
        """

        service_descriptor = metadata_service_pb2.DESCRIPTOR.services_by_name["MetadataService"]
        request_serializer = None
        response_deserializer = None
        for method in service_descriptor.methods:
            if method.name != method_name:
                continue

            if method.input_type.name == "Empty":
                request_serializer = Empty.SerializeToString
            else:
                request_serializer = getattr(metadata_service_pb2, method.input_type.name).SerializeToString
            if method.output_type.name == "Empty":
                response_deserializer = Empty.FromString
            else:
                response_deserializer = getattr(metadata_service_pb2, method.output_type.name).FromString

        assert (
            request_serializer is not None and response_deserializer is not None
        ), f"No request_serializer or response_deserializer match for {method_name}"
        return request_serializer, response_deserializer



import re
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

import pandas
import pendulum
import sqlparse

from tecton_proto.common import aggregation_function_pb2 as afpb
from tecton_proto.common import column_type_pb2
from tecton_proto.data.batch_data_source_pb2 import BatchDataSource
from tecton_proto.data.feature_view_pb2 import FeatureView as FeatureViewProto
from tecton_proto.data.feature_view_pb2 import NewTemporalAggregate
from tecton_spark import conf
from tecton_spark import errors
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_spark.feature_set_config import FeatureSetConfig
from tecton_spark.materialization_context import BaseMaterializationContext
from tecton_spark.materialization_context import BoundMaterializationContext
from tecton_spark.snowflake.errors import TectonSnowflakeNotImplementedError
from tecton_spark.snowflake.pipeline_helper import pipeline_to_df_with_input
from tecton_spark.snowflake.pipeline_helper import pipeline_to_sql_string
from tecton_spark.snowflake.templates_utils import load_template

TEMP_SPINE_TABLE_NAME = "_TEMP_SPINE_TABLE_FROM_DF"

FULL_AGGREGATION_TEMPLATE = load_template("run_full_aggregation.sql")
HISTORICAL_FEATURES_TEMPLATE = load_template("historical_features.sql")
MATERIALIZATION_TILE_TEMPLATE = load_template("materialization_tile.sql")
MATERIALIZED_FEATURE_VIEW_TEMPLATE = load_template("materialized_feature_view.sql")
ONLINE_STORE_COPIER_TEMPLATE = load_template("online_store_copier.sql")
PARTIAL_AGGREGATION_TEMPLATE = load_template("run_partial_aggregation.sql")
TIME_LIMIT_TEMPLATE = load_template("time_limit.sql")

# TODO(TEC-6204): Last and LastN are not currently supported
#
# Map of proto function type -> set of (output column prefix, snowflake function name)
AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: {("SUM", "SUM")},
    afpb.AGGREGATION_FUNCTION_MIN: {("MIN", "MIN")},
    afpb.AGGREGATION_FUNCTION_MAX: {("MAX", "MAX")},
    afpb.AGGREGATION_FUNCTION_COUNT: {("COUNT", "COUNT")},
    afpb.AGGREGATION_FUNCTION_MEAN: {("COUNT", "COUNT"), ("MEAN", "AVG")},
}


def _format_sql(sql_str: str) -> str:
    return sqlparse.format(sql_str, reindent=True, keyword_case="upper")


@dataclass
class _FeatureSetItemInput:
    """A simplified version of FeatureSetItem which is passed to the SQL template."""

    name: str
    timestamp_key: str
    join_keys: Dict[str, str]
    features: List[str]
    sql: str
    aggregation: Optional[NewTemporalAggregate]
    ttl_seconds: Optional[int]


def get_historical_features_sql(
    spine_sql: Optional[str],
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    include_feature_view_timestamp_columns: bool = False,
    start_time: datetime = None,
    end_time: datetime = None,
    session: "snowflake.snowpark.Session" = None,
) -> str:
    feature_set_items = feature_set_config._get_feature_definitions_and_join_configs()
    input_items = []
    if spine_sql is None:
        # Only feature view is supported when the spine is not provided.
        # Feature service should always provide the spine.
        # SDK methods should never fail this check
        assert len(feature_set_items) == 1

    # Get a list of all the join keys in the spine.
    spine_keys = {}
    for item in feature_set_items:
        fd = item.feature_definition

        if not fd.is_on_demand:
            join_keys = {key: value for key, value in item.join_keys}
            spine_keys.update(join_keys)

    for item in feature_set_items:
        fd = item.feature_definition
        if fd.is_on_demand and not conf.get_bool("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
            raise TectonSnowflakeNotImplementedError("On-demand features are only supported with Snowpark enabled")
        feature_view = fd.fv
        # Change the feature view name if it's for internal udf use.
        if item.namespace.startswith("_udf_internal"):
            name = item.namespace.upper()
        else:
            name = fd.name

        if not fd.is_on_demand:
            join_keys = {key: value for key, value in item.join_keys}
            features = [
                col_name
                for col_name in fd.view_schema.column_names()
                if col_name not in (list(join_keys.keys()) + [fd.timestamp_key])
            ]
            if len(fd.online_serving_index.join_keys) != len(fd.join_keys):
                raise TectonSnowflakeNotImplementedError("Wildcard is not supported for Snowflake")
            if start_time is None or (
                fd.feature_start_timestamp is not None and start_time < fd.feature_start_timestamp
            ):
                start_time = fd.feature_start_timestamp

            raw_data_start_time = start_time
            raw_dat_end_time = end_time
            if fd.is_temporal_aggregate and raw_data_start_time is not None:
                # Account for final aggregation needing aggregation window prior to earliest timestamp
                max_aggregation_window = fd.max_aggregation_window
                raw_data_start_time = raw_data_start_time - max_aggregation_window.ToTimedelta()
            sql_str = generate_run_batch_sql(
                feature_definition=fd,
                feature_start_time=raw_data_start_time,
                feature_end_time=raw_dat_end_time,
                spine=spine_sql,
                spine_timestamp_key=timestamp_key,
                session=session,
            )
            input_items.append(
                _FeatureSetItemInput(
                    name=name,
                    timestamp_key=fd.timestamp_key,
                    join_keys=join_keys,
                    features=features,
                    sql=sql_str,
                    aggregation=(
                        feature_view.temporal_aggregate if feature_view.HasField("temporal_aggregate") else None
                    ),
                    ttl_seconds=(int(fd.serving_ttl.total_seconds()) if fd.is_temporal else None),
                )
            )
    if spine_sql is not None:
        sql_str = HISTORICAL_FEATURES_TEMPLATE.render(
            feature_set_items=input_items,
            spine_timestamp_key=timestamp_key,
            spine_sql=spine_sql,
            include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
            spine_keys=list(spine_keys),
        )
    if start_time is not None or end_time is not None:
        timestamp = timestamp_key if spine_sql is not None else fd.timestamp_key
        sql_str = TIME_LIMIT_TEMPLATE.render(
            source=sql_str, timestamp_key=timestamp, start_time=start_time, end_time=end_time
        )

    return _format_sql(sql_str)


def get_historical_features(
    spine: Optional[Union[pandas.DataFrame, str]],
    connection: "snowflake.connector.Connection",
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    include_feature_view_timestamp_columns: bool = False,
    start_time: datetime = None,
    end_time: datetime = None,
) -> pandas.DataFrame:
    cur = connection.cursor()
    spine_sql = None
    if isinstance(spine, str):
        spine_sql = spine
    elif isinstance(spine, pandas.DataFrame):
        spine_sql = generate_sql_table_from_pandas_df(df=spine, table_name=TEMP_SPINE_TABLE_NAME, connection=connection)

    sql_str = get_historical_features_sql(
        spine_sql=spine_sql,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        start_time=start_time,
        end_time=end_time,
    )
    cur.execute(sql_str)
    return cur.fetch_pandas_all()


def generate_sql_table_from_pandas_df(
    df: pandas.DataFrame,
    table_name: str,
    session: "snowflake.snowpark.Session" = None,
    connection: "snowflake.connector.Connection" = None,
) -> str:
    """Generate a TABLE from pandas.DataFrame. Returns the sql query to select * from the table"""
    if session is None and connection is None:
        raise ValueError("Either session or connection must be provided")

    if session is not None:
        session.sql(f"DROP TABLE IF EXISTS {table_name}").collect()
        session.write_pandas(df, table_name, auto_create_table=True, create_temp_table=True)
        return f"SELECT * FROM {table_name}"

    if connection is not None:
        from snowflake.connector.pandas_tools import write_pandas

        # Get the SQL that would be generated by the create table statement
        create_table_sql = pandas.io.sql.get_schema(df, table_name)

        # Replace the `CREATE TABLE` with `CREATE OR REPLACE TEMPORARY TABLE`
        create_tmp_table_sql = re.sub("^(CREATE TABLE)?", "CREATE OR REPLACE TEMPORARY TABLE", create_table_sql)
        cur = connection.cursor()
        cur.execute(create_tmp_table_sql)
        write_pandas(conn=connection, df=df, table_name=table_name)
        return f"SELECT * FROM {table_name}"


def validate_spine_dataframe(spine_df: "snowflake.snowpark.DataFrame", timestamp_key: str, join_keys: List[str]):
    if timestamp_key not in spine_df.columns:
        raise errors.TectonValidationError(
            f"Expected to find '{timestamp_key}' among available spine columns: '{', '.join(spine_df.columns)}'."
        )
    for field in spine_df.schema.fields:
        if field.name == timestamp_key and field.datatype.type_name != "Timestamp":
            raise errors.TectonValidationError(
                f"Invalid type of timestamp_key column in the given spine. Expected Timestamp, got {field.datatype}"
            )
    for key in join_keys:
        if key not in spine_df.columns:
            raise errors.TectonValidationError(
                f"Expected to find '{key}' among available spine columns: '{', '.join(spine_df.columns)}'."
            )


def get_historical_features_with_snowpark(
    spine: Union[pandas.DataFrame, str, "snowflake.snowpark.DataFrame"],
    session: "snowflake.snowpark.Session",
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    include_feature_view_timestamp_columns: bool = False,
    start_time: datetime = None,
    end_time: datetime = None,
    entities: "snowflake.snowpark.DataFrame" = None,
) -> "snowflake.snowpark.DataFrame":
    from snowflake.snowpark import DataFrame

    spine_sql = None
    if isinstance(spine, str):
        spine_sql = spine
    elif isinstance(spine, DataFrame):
        spine.write.save_as_table(TEMP_SPINE_TABLE_NAME, mode="overwrite", create_temp_table=True)
        spine_sql = f"SELECT * FROM {TEMP_SPINE_TABLE_NAME}"
    elif isinstance(spine, pandas.DataFrame):
        spine_sql = generate_sql_table_from_pandas_df(df=spine, table_name=TEMP_SPINE_TABLE_NAME, session=session)

    if spine is not None:
        if timestamp_key is None:
            schema = session.sql(spine_sql).schema
            timestamp_cols = [field.name for field in schema.fields if field.datatype.type_name == "Timestamp"]

            if len(timestamp_cols) > 1 or len(timestamp_cols) == 0:
                raise errors.TectonValidationError(
                    f"Could not infer timestamp keys from {schema}; please specify explicitly"
                )
            timestamp_key = timestamp_cols[0]

        join_keys = [join_key for fd in feature_set_config.feature_definitions for join_key in fd.join_keys]
        validate_spine_dataframe(session.sql(spine_sql), timestamp_key, join_keys=join_keys)

    sql_str = get_historical_features_sql(
        spine_sql=spine_sql,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        start_time=start_time,
        end_time=end_time,
        session=session,
    )
    output_df = session.sql(sql_str)

    # Apply ODFV to the spine.
    for item in feature_set_config._get_feature_definitions_and_join_configs():
        fd = item.feature_definition
        if fd.is_on_demand:
            view_schema = fd.view_schema
            schema_dict = {col_name: view_schema.spark_type(col_name) for col_name in view_schema.column_names()}
            output_df = pipeline_to_df_with_input(
                session=session,
                input_df=output_df,
                pipeline=fd.pipeline,
                transformations=fd.transformations,
                output_schema=schema_dict,
                name=fd.name,
                fv_id=fd.id,
            )
    columns_to_drop = [column for column in output_df.columns if "_UDF_INTERNAL_" in column]
    if len(columns_to_drop) > 0:
        output_df = output_df.drop(*columns_to_drop)
    if entities is not None:
        column_names = [field.name for field in entities.schema.fields]
        # Do an inner join on the entities to filter out rows that don't have a matching entity
        output_df = entities.join(right=output_df, using_columns=column_names, join_type="inner")
    return output_df


def generate_run_batch_sql(
    feature_definition: FeatureDefinition,
    # start is inclusive and end is exclusive
    feature_start_time: Optional[datetime] = None,
    feature_end_time: Optional[datetime] = None,
    aggregation_level: str = "full",
    # If spine is provided, it will be used to join with the output results.
    # Currently only work with full aggregation.
    spine: Optional[str] = None,
    spine_timestamp_key: Optional[str] = None,
    spine_keys: Optional[List[str]] = None,
    mock_sql_inputs: Optional[Dict[str, str]] = None,
    materialization_context: Optional[BaseMaterializationContext] = None,
    session: "snowflake.snowpark.Session" = None,
) -> str:
    feature_view = feature_definition.feature_view_proto

    # Set a default materilization_context if not provided.
    # This is following the same logic as spark.
    if materialization_context is None:
        materialization_context = BoundMaterializationContext._create_internal(
            feature_start_time or pendulum.from_timestamp(0, pendulum.tz.UTC),
            feature_end_time or pendulum.datetime(2100, 1, 1),
            pendulum.Duration(seconds=feature_view.materialization_params.schedule_interval.ToSeconds()),
        )
    if feature_definition.is_incremental_backfill:
        materialized_sql = TIME_LIMIT_TEMPLATE.render(
            source=f"SELECT * FROM {feature_view.snowflake_data.snowflake_view_name}",
            timestamp_key=feature_view.timestamp_key,
            start_time=feature_start_time,
            end_time=feature_end_time,
        )
    else:
        pipeline_sql = pipeline_to_sql_string(
            pipeline=feature_definition.pipeline,
            data_sources=feature_definition.data_sources,
            transformations=feature_definition.transformations,
            materialization_context=materialization_context,
            mock_sql_inputs=mock_sql_inputs,
            session=session,
        )
        materialized_sql = get_materialization_query(
            feature_view=feature_view,
            feature_start_time=feature_start_time,
            feature_end_time=feature_end_time,
            source=pipeline_sql,
        )
    if feature_view.HasField("temporal_aggregate"):
        if aggregation_level == "full":
            aggregated_sql_str = FULL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=list(feature_view.join_keys),
                aggregation=feature_view.temporal_aggregate,
                timestamp_key=feature_view.timestamp_key,
                name=feature_view.fco_metadata.name,
                spine=spine,
                spine_timestamp_key=spine_timestamp_key,
                spine_keys=spine_keys,
            )
            return _format_sql(aggregated_sql_str)
        elif aggregation_level == "partial":
            # Rename the output columns, and add tile start/end time columns
            partial_aggregated_sql_str = PARTIAL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=list(feature_view.join_keys),
                aggregations=_get_feature_view_aggregations(feature_view),
                slide_interval=feature_view.temporal_aggregate.slide_interval,
                slide_interval_string=feature_view.temporal_aggregate.slide_interval_string,
                timestamp_key=feature_view.timestamp_key,
            )
            return _format_sql(partial_aggregated_sql_str)
        elif aggregation_level == "disabled":
            sql_str = TIME_LIMIT_TEMPLATE.render(
                source=pipeline_sql,
                timestamp_key=feature_view.timestamp_key,
                start_time=feature_start_time,
                end_time=feature_end_time,
            )
            return _format_sql(sql_str)
        else:
            raise ValueError(f"Unsupported aggregation level: {aggregation_level}")

    else:
        return _format_sql(materialized_sql)


# By default Snowflake unloads numerical types as byte arrays because they have higher precision than available Parquet
# types. We could decode these in the OnlineStoreCopier, but since we will be downcasting them to the precision of
# Parquet types we might as well do it now.
#
# Snowflake docs for how casting interacts with Parquet types:
# https://docs.snowflake.com/en/user-guide/data-unload-considerations.html#explicitly-converting-numeric-columns-to-parquet-data-types
COPY_CASTS = {
    column_type_pb2.COLUMN_TYPE_UNKNOWN: None,
    column_type_pb2.COLUMN_TYPE_INT64: "BIGINT",
    column_type_pb2.COLUMN_TYPE_DOUBLE: "DOUBLE",
    column_type_pb2.COLUMN_TYPE_STRING: None,
    column_type_pb2.COLUMN_TYPE_BOOL: None,
    column_type_pb2.COLUMN_TYPE_STRING_ARRAY: None,
    # TODO(TEC-8501): Probably need to do something here
    column_type_pb2.COLUMN_TYPE_INT64_ARRAY: None,
    column_type_pb2.COLUMN_TYPE_DOUBLE_ARRAY: None,
    column_type_pb2.COLUMN_TYPE_FLOAT_ARRAY: None,
}


def get_materialization_query(
    feature_view: FeatureViewProto,
    source: str,
    # start is inclusive and end is exclusive
    feature_start_time: datetime = None,
    feature_end_time: datetime = None,
):
    """Returns a SQL query for time-limited materialization.

    Does not include a terminating `;` or any COPY or INSERT statements."""
    if feature_view.HasField("temporal_aggregate"):
        source = MATERIALIZATION_TILE_TEMPLATE.render(
            source=source,
            join_keys=list(feature_view.join_keys),
            aggregations=_get_feature_view_aggregations(feature_view),
            slide_interval=feature_view.temporal_aggregate.slide_interval,
            timestamp_key=feature_view.timestamp_key,
        )
    return _format_sql(
        TIME_LIMIT_TEMPLATE.render(
            source=source,
            timestamp_key=feature_view.timestamp_key,
            start_time=feature_start_time,
            end_time=feature_end_time,
        )
    )


def get_materialization_copy_sql(
    feature_definition: FeatureDefinition,
    # start is inclusive and end is exclusive
    time_limits: pendulum.Period,
    destination_stage: str,
    destination_table: Optional[str],
    materialization_id: str,
    session: "snowflake.snowpark.Session" = None,
):
    """Returns a SQL query for a COPY INTO an destination_stage for materialization.

    Additionally INSERTs into destination_table for manually-materialized FeatureViews."""
    feature_view = feature_definition.feature_view_proto

    materialization_context = BoundMaterializationContext._create_internal(
        time_limits.start,
        time_limits.end,
        pendulum.Duration(seconds=feature_view.materialization_params.schedule_interval.ToSeconds()),
    )

    source = pipeline_to_sql_string(
        pipeline=feature_definition.pipeline,
        data_sources=feature_definition.data_sources,
        transformations=feature_definition.transformations,
        materialization_context=materialization_context,
        session=session,
    )
    query = get_materialization_query(
        source=source, feature_view=feature_view, feature_start_time=time_limits.start, feature_end_time=time_limits.end
    )
    common_context = dict(
        source=query,
        destination_stage=destination_stage,
        materialization_schema=feature_view.schemas.materialization_schema,
        materialization_id=materialization_id,
        cast_types=COPY_CASTS,
    )
    if destination_table:
        script_sql = MATERIALIZED_FEATURE_VIEW_TEMPLATE.render(destination_table=destination_table, **common_context)
        sql = "\n".join(("EXECUTE IMMEDIATE", "$$", _format_sql(script_sql), "$$;"))
    else:
        sql = ONLINE_STORE_COPIER_TEMPLATE.render(**common_context)
    return _format_sql(sql)


def get_dataframe_for_data_source(
    session: "snowflake.snowpark.Session",
    data_source: BatchDataSource,
    start_time: datetime = None,
    end_time: datetime = None,
) -> "snowflake.snowpark.DataFrame":
    snowflake_args = data_source.snowflake.snowflakeArgs
    source = f"{snowflake_args.database}.{snowflake_args.schema}.{snowflake_args.table}"

    if (start_time is not None or end_time is not None) and snowflake_args.timestamp_key is None:
        raise ValueError(
            "Filtering by start_time or end_time requires the timestamp_key parameter to be set on this data source"
        )
    sql_str = TIME_LIMIT_TEMPLATE.render(
        source=source,
        timestamp_key=snowflake_args.timestamp_key,
        start_time=start_time,
        end_time=end_time,
    )
    return session.sql(sql_str)


def _get_feature_view_aggregations(feature_view: FeatureViewProto) -> Dict[str, Set[str]]:
    aggregations = defaultdict(set)
    for feature in feature_view.temporal_aggregate.features:
        aggregate_function = AGGREGATION_PLANS[feature.function]
        if not aggregate_function:
            raise TectonSnowflakeNotImplementedError(
                f"Unsupported aggregation function {feature.function} in snowflake pipeline"
            )
        aggregations[feature.input_feature_name].update(aggregate_function)

    # Need to order the functions for deterministic results.
    for key, value in aggregations.items():
        aggregations[key] = sorted(value)

    return aggregations



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/version_constraints.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from tecton_proto.common import framework_version_pb2 as tecton__proto_dot_common_dot_framework__version__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/version_constraints.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n+tecton_proto/args/version_constraints.proto\x12\x11tecton_proto.args\x1a google/protobuf/descriptor.proto\x1a+tecton_proto/common/framework_version.proto\"\xad\x01\n\x1b\x46rameworkVersionConstraints\x12\x46\n\x0bmin_version\x18\x01 \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionR\nminVersion\x12\x46\n\x0bmax_version\x18\x02 \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionR\nmaxVersion:{\n\x11\x66ramework_version\x12\x1d.google.protobuf.FieldOptions\x18\xd0\x0f \x01(\x0b\x32..tecton_proto.args.FrameworkVersionConstraintsR\x10\x66rameworkVersionB\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,tecton__proto_dot_common_dot_framework__version__pb2.DESCRIPTOR,])


FRAMEWORK_VERSION_FIELD_NUMBER = 2000
framework_version = _descriptor.FieldDescriptor(
  name='framework_version', full_name='tecton_proto.args.framework_version', index=0,
  number=2000, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='frameworkVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)


_FRAMEWORKVERSIONCONSTRAINTS = _descriptor.Descriptor(
  name='FrameworkVersionConstraints',
  full_name='tecton_proto.args.FrameworkVersionConstraints',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_version', full_name='tecton_proto.args.FrameworkVersionConstraints.min_version', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='minVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_version', full_name='tecton_proto.args.FrameworkVersionConstraints.max_version', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='maxVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=146,
  serialized_end=319,
)

_FRAMEWORKVERSIONCONSTRAINTS.fields_by_name['min_version'].enum_type = tecton__proto_dot_common_dot_framework__version__pb2._FRAMEWORKVERSION
_FRAMEWORKVERSIONCONSTRAINTS.fields_by_name['max_version'].enum_type = tecton__proto_dot_common_dot_framework__version__pb2._FRAMEWORKVERSION
DESCRIPTOR.message_types_by_name['FrameworkVersionConstraints'] = _FRAMEWORKVERSIONCONSTRAINTS
DESCRIPTOR.extensions_by_name['framework_version'] = framework_version
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FrameworkVersionConstraints = _reflection.GeneratedProtocolMessageType('FrameworkVersionConstraints', (_message.Message,), {
  'DESCRIPTOR' : _FRAMEWORKVERSIONCONSTRAINTS,
  '__module__' : 'tecton_proto.args.version_constraints_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.FrameworkVersionConstraints)
  })
_sym_db.RegisterMessage(FrameworkVersionConstraints)

framework_version.message_type = _FRAMEWORKVERSIONCONSTRAINTS
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(framework_version)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import collections

try:
    import grpc
except ImportError:
    raise ImportError("Please install grpcio")

from google.protobuf.empty_pb2 import Empty

from tecton_spark import conf

from tecton._internals.metadata_service_impl import base_stub
from tecton._internals.metadata_service_impl import error_lib
from tecton._internals.metadata_service_impl import request_lib
from tecton_proto.metadataservice import metadata_service_pb2
from collections import defaultdict
from tecton._internals.metadata_service_impl.response import MDSResponse


def _get_host_port() -> str:
    return conf.get_or_raise("METADATA_SERVICE")


class MetadataUnaryUnaryWrapper(object):
    """
    Adding headers from server based on this
    https://github.com/grpc/grpc/blob/master/src/python/grpcio/grpc/_interceptor.py
    """

    def __init__(self, fn):
        self.fn = fn

    def __call__(self, request):
        response_proto, call = self.fn.with_call(request=request)
        return MDSResponse(response_proto, defaultdict(str, dict(call.initial_metadata())))


class MetadataServiceStub(base_stub.BaseStub):
    # Due to https://github.com/stackb/rules_proto/issues/113 generating GRPC
    # classes using protoc is not working. This is manually recreating what
    # protoc does for the client side of GRPC services. It only supports unary-unary.
    # If anything else is needed hopefully that bug has been fixed.
    def __init__(self, channel):
        service_descriptor = metadata_service_pb2.DESCRIPTOR.services_by_name["MetadataService"]
        for method in service_descriptor.methods:
            if method.input_type.name == "Empty":
                request_serializer = Empty.SerializeToString
            else:
                request_serializer = getattr(metadata_service_pb2, method.input_type.name).SerializeToString
            if method.output_type.name == "Empty":
                response_deserializer = Empty.FromString
            else:
                response_deserializer = getattr(metadata_service_pb2, method.output_type.name).FromString
            fn = channel.unary_unary(
                f"/{service_descriptor.full_name}/{method.name}",
                request_serializer=request_serializer,
                response_deserializer=response_deserializer,
            )
            self._channel = channel
            setattr(self, method.name, MetadataUnaryUnaryWrapper(fn))

    def close(self):
        self._channel.close()


class _ClientCallDetails(
    collections.namedtuple("_ClientCallDetails", ("method", "timeout", "metadata", "credentials")),
    grpc.ClientCallDetails,
):
    pass


class MetadataServiceInterceptor(grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor):
    """
    Adding headers based on an a example in
    https://github.com/grpc/grpc/blob/master/examples/python/interceptors/headers/header_manipulator_client_interceptor.py
    """

    def __init__(self):
        pass

    @staticmethod
    def _intercept_call(continuation, client_call_details, request_or_iterator):
        metadata = [(k, v) for k, v in request_lib.request_headers().items()]

        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata, client_call_details.credentials
        )

        response = continuation(client_call_details, request_or_iterator)

        e = response.exception()
        if not e:
            return response

        if isinstance(e, grpc.RpcError):
            error_lib.raise_for_grpc_status(
                status_code=e.code().value[0], details=e.details(), host_url=_get_host_port()
            )

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator)
